

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
//_/_/
//_/_/  This file is part of CMSDK version 2.0.2.180713 (13 July 2018)
//_/_/  
//_/_/  developed by
//_/_/  
//_/_/  Communicative Machines Ltd. 
//_/_/    http://cmlabs.com/
//_/_/  Copyright (c) 2018
//_/_/  
//_/_/  in collaboration with 
//_/_/  
//_/_/  Icelandic Institute for Intelligent Machines ses
//_/_/    Menntavegur 1
//_/_/    101 Reykjavik
//_/_/    ICELAND
//_/_/    http://www.iiim.is/
//_/_/
//_/_/  and is released under the following
//_/_/  
//_/_/  EXTENDED BSD License
//_/_/
//_/_/  Redistribution and use in source and binary forms, with or without
//_/_/  modification, is permitted provided that the following conditions 
//_/_/  are met:
//_/_/
//_/_/  - Redistributions of source code must retain the above copyright 
//_/_/    and collaboration notice, this list of conditions and the following 
//_/_/    disclaimer.
//_/_/
//_/_/  - Redistributions in binary form must reproduce the above copyright 
//_/_/    notice, this list of conditions and the following disclaimer in the 
//_/_/    documentation and/or other materials provided with the distribution.
//_/_/
//_/_/  - Neither the name of its copyright holders nor the names of its 
//_/_/    contributors may be used to endorse or promote products derived from 
//_/_/    this software without specific prior written permission.
//_/_/
//_/_/  - CADIA Clause: The license granted in and to the software under this 
//_/_/    agreement is a limited-use license. The software may not be used in 
//_/_/    furtherance of: (i) intentionally causing bodily injury or severe 
//_/_/    emotional harm to any person; (ii) invading the personal privacy or 
//_/_/    violating the human rights of any person; or (iii) committing or 
//_/_/    preparing for any act of war. 
//_/_/
//_/_/  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//_/_/  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//_/_/  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
//_/_/  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
//_/_/  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
//_/_/  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
//_/_/  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//_/_/  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//_/_/  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//_/_/  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//_/_/  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//_/_/
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
#include "MovingAverage.h"

namespace cmlabs {

Stats::Stats() {
}

Stats::~Stats() {
	mutex.enter();
	entries.clear();
	mutex.leave();
}

bool Stats::clear() {
	mutex.enter();
	entries.clear();
	mutex.leave();
	return true;
}

bool Stats::add(double val) {
	mutex.enter();
	entries.push_back(val);
	mutex.leave();
	return true;
}

double Stats::getSum() {
	double result = 0;
	mutex.enter();
	std::vector<double>::iterator i, e;
	for (i=entries.begin(), e=entries.end(); i!=e; i++)
		result += *i;
	mutex.leave();
	return result;
}

double Stats::getAverage() {
	double result;
	mutex.enter();
	result = getSum()/entries.size();
	mutex.leave();
	return result;
}

double Stats::getVariance() {
	double variance, average, size;
	mutex.enter();
	if ( (size = (double)entries.size()) <= 1) {
		mutex.leave();
		return 0;
	}
	variance = 0;
	average = getSum()/size;
	std::vector<double>::iterator i, e;
	for (i=entries.begin(), e=entries.end(); i!=e; i++)
		variance += (*i-average)*(*i-average);
	variance /= size;
	mutex.leave();
	return variance;
}

double Stats::getMedian() {
	double result;
	uint32 size, half;
	mutex.enter();
	size = (uint32)entries.size();
	half = size / 2;
	std::sort(entries.begin(), entries.end());
	if (size % 2)
		result = entries[half-1];
	else
		result = (entries[half-1] + entries[half]) / 2.0;
	mutex.leave();
	return result;
}

double Stats::getStdDev() {
	return sqrt(getVariance());
}

uint32 Stats::getCount() {
	uint32 result;
	mutex.enter();
	result = (uint32)entries.size();
	mutex.leave();
	return result;
}


bool Stats::UnitTest() {
	Stats stats;
	stats.add(1);
	stats.add(3);
	stats.add(10);
	stats.add(-5);
	stats.add(8);
	stats.add(100);
	stats.add(10);
	stats.add(1);
	
	if (stats.getAverage() != 16)
		return false;
	if (stats.getMedian() != 5.5)
		return false;
	if (stats.getVariance() != 1031.5)
		return false;
	if (abs(stats.getStdDev() - 32.11697) > 0.0001)
		return false;
	
	return true;
}





} // namespace cmlabs
