

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
//_/_/
//_/_/  This file is part of PSYCLONE version 2.0.2.180713 (13 July 2018)
//_/_/  
//_/_/  developed by
//_/_/  
//_/_/  Communicative Machines Ltd. 
//_/_/    http://cmlabs.com/
//_/_/  Copyright (c) 2018
//_/_/  
//_/_/  in collaboration with 
//_/_/  
//_/_/  Icelandic Institute for Intelligent Machines ses
//_/_/    Menntavegur 1
//_/_/    101 Reykjavik
//_/_/    ICELAND
//_/_/    http://www.iiim.is/
//_/_/
//_/_/  PSYCLONE 2.0 is released under the GNU Lesser General Public License 
//_/_/  (LGPL) version 3.0, and is subject to the following
//_/_/  
//_/_/  - CADIA Clause: The license granted in and to the software under this 
//_/_/    agreement is a limited-use license. The software may not be used in 
//_/_/    furtherance of: (i) intentionally causing bodily injury or severe 
//_/_/    emotional harm to any person; (ii) invading the personal privacy or 
//_/_/    violating the human rights of any person; or (iii) committing or 
//_/_/    preparing for any act of war. 
//_/_/    
//_/_/  A commercial license for PSYCLONE 2.0 is also available to 
//_/_/  any interested party from the copyright holder.
//_/_/  
//_/_/  You can redistribute this copy of PSYCLONE 2.0 and/or modify it under 
//_/_/  the terms of the GNU General Public License as published by the Free 
//_/_/  Software Foundation, either version 3 of the License, or (at your 
//_/_/  option) any later version.
//_/_/  
//_/_/  You should have received a copy of the GNU General Public License
//_/_/  and the GNU Lesser General Public License along with PSYCLONE 2.0.  
//_/_/  If not, see <http://www.gnu.org/licenses/>.
//_/_/  
//_/_/  PSYCLONE 2.0 is distributed in the hope that it will be useful,
//_/_/  but WITHOUT ANY WARRANTY; without even the implied warranty of
//_/_/  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//_/_/  GNU General Public License for more details.
//_/_/
//_/_/  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//_/_/  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//_/_/  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
//_/_/  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
//_/_/  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
//_/_/  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
//_/_/  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//_/_/  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//_/_/  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//_/_/  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//_/_/  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//_/_/
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
#if !defined(_CONSOLE_H_)
#define _CONSOLE_H_

#pragma once

#include "MemoryManager.h"
#include "Utils.h"
#include "PsyTime.h"
#include "NetworkManager.h"

#define STDIN 0

namespace	cmlabs{

// 0-4 already defined in Utils.h
//#define LOG_NOTOPIC		0
//#define LOG_SYSTEM		1
//#define LOG_NETWORK		2
//#define LOG_MEMORY		3
//#define LOG_PROCESS		4
#define LOG_SYNC			5
#define LOG_NODE			6
#define LOG_SPACE			7
#define LOG_COMPONENT		8
#define LOG_SUBSCRIPTIONS	9
#define LOG_TRIGGERS		10
#define LOG_TIMING			11
#define SUBJECT_COUNT		11

static const char *LogSubjects[] = {
	"",
	"System",
	"Network",
	"Memory",
	"Process",
	"Sync",
	"Node",
	"Space",
	"Component",
	"Subscriptions",
	"Triggers",
	"Timing"
};
static const char *VerboseSubjects[] = {
	"verbose",
	"verbose-system",
	"verbose-network",
	"verbose-memory",
	"verbose-process",
	"verbose-sync",
	"verbose-node",
	"verbose-space",
	"verbose-component",
	"verbose-subscriptions",
	"verbose-triggers",
	"verbose-timing"
};
static const char *DebugSubjects[] = { 
	"debug",
	"debug-system",
	"debug-network",
	"debug-memory",
	"debug-process",
	"debug-sync",
	"debug-node",
	"debug-space",
	"debug-component",
	"debug-subscriptions",
	"debug-triggers",
	"debug-timing"
};

struct ChannelProfile {
	uint32	id;
	uint64	conid;
	uint64	lastinput;
	uint8	logLevelVerbose[LOG_MAXCOUNT];
	uint8	logLevelDebug[LOG_MAXCOUNT];
};


class Node;

class Console : public Runnable, public LogReceiver, public NetworkReceiver
{
public:
	friend THREAD_RET THREAD_FUNCTION_CALL ConsoleRun(THREAD_ARG arg);

	// ***************** Constructors *****************
	Console();
	virtual ~Console();

	// ***************** Initialisation *****************
	bool init(Node* parent, NetworkManager* channel, const char* welcomeText = NULL);

	bool connectClient(uint64 conid);
	bool disconnectClient(uint64 conid);
	uint32 getClientID(uint64 conid);

	bool receiveTelnetLine(TelnetLine* line, NetworkChannel* channel, uint64 conid);

	bool logEntry(LogEntry* entry);
	bool processInput(uint32 id, const char* text);
	bool processSignal(uint32 id, int sig);

	bool printToClient(uint32 id, const char* format, ...);

	bool setLogLevelDebug(uint32 id, uint8 level);
	bool setLogLevelDebug(uint32 id, uint8 subject, uint8 level);
	bool setLogLevelVerbose(uint32 id, uint8 level);
	bool setLogLevelVerbose(uint32 id, uint8 subject, uint8 level);

protected:
	uint64 lastBreak;
	uint8 breakCount;
	uint32 lastID;
	std::map<uint32, ChannelProfile> clients;
	char* welcomeText;

	NetworkManager* networkManager;
	Node* node;

	// ***************** Running *****************
	bool run();
};

THREAD_RET THREAD_FUNCTION_CALL ConsoleRun(THREAD_ARG arg);

}

#include "Node.h"

#endif // _CONSOLE_H_
